/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <genht/hash.h>
#include "libfawk.h"
#include "builtin.h"
#include "str.h"
#include "execute.h"

FAWK_API void fawk_close_include(fawk_ctx_t *ctx, fawk_src_t *src)
{
	if (src->fn == NULL) return;
	if (ctx->parser.include != NULL) ctx->parser.include(ctx, src, 0, NULL);
	fawk_free(ctx, src->fn); src->fn = NULL;
}

FAWK_API void fawk_init(fawk_ctx_t *ctx)
{
	memset(ctx, 0, sizeof(fawk_ctx_t));
	htsp_init(&ctx->symtab, strhash, strkeyeq);
	fawk_builtin_init(ctx);
	ctx->parser.isp = &ctx->parser.include_stack[0];
}

FAWK_API void fawk_uninit(fawk_ctx_t *ctx)
{
	fawk_src_t *src;
	size_t n;
	htsp_entry_t *e;

	/* free the stack */
	fawk_reset(ctx);
	for(n = 0; n < ctx->stack.used; n++)
		fawk_free(ctx, ctx->stack.page[n]);
	fawk_free(ctx, ctx->stack.page);

	/* free globals */
	for (e = htsp_first(&ctx->symtab); e; e = htsp_next(&ctx->symtab, e)) {
		fawk_free(ctx, e->key);
		fawk_cell_free(ctx, e->value);
		fawk_free(ctx, e->value);
	}

	htsp_uninit(&ctx->symtab);

	/* free the bytecode and attached constants */
	for(n = 0; n < ctx->code.used; n++) {
		switch(ctx->code.code[n].type) {
			case FAWKC_INS: case FAWKC_NUM: case FAWKC_CSTR: break;
			case FAWKC_STR: fawk_str_free(ctx, ctx->code.code[n].data.str); break;
			case FAWKC_SYMREF:
				fawk_free(ctx, ctx->code.code[n].data.symref->idx);
				fawk_free(ctx, ctx->code.code[n].data.symref);
				break;
		}
	}
	fawk_free(ctx, ctx->code.code);
	fawk_free(ctx, ctx->parser.buff);

	/* in case of an error exit, there may be files left open */
	for(src = ctx->parser.include_stack; src <= ctx->parser.isp; src++) fawk_close_include(ctx, src);

	FAWK_PKG_CALL(ctx, p->uninit_cb, (p, ctx)); FAWK_PKG_CALL(ctx, free, (p));
}

FAWK_API char *fawk_strdup(fawk_ctx_t *ctx, const char *s)
{
	size_t l = strlen(s);
	char *ret = fawk_malloc(ctx, l+1);
	if (ret != NULL) memcpy(ret, s, l+1);
	return ret;
}

FAWK_API void fawk_errbuff(fawk_ctx_t *ctx, size_t len)
{
	if (len > ctx->errbuff_alloced) {
		fawk_free(ctx, ctx->errbuff);
		ctx->errbuff_alloced = len;
		ctx->errbuff = fawk_malloc(ctx, len);
	}
	if (ctx->errbuff != NULL) *ctx->errbuff = '\0';
}

FAWK_API void fawk_dump_cell(fawk_cell_t *cell, int verbose)
{
	switch(cell->type) {
		case FAWK_NUM:    if (verbose) printf("NUM:{"FAWK_NUM_PRINTF_FMT"}", cell->data.num); else printf(FAWK_NUM_PRINTF_FMT, cell->data.num); return;
		case FAWK_STR:    if (verbose) printf("STR:{'%s' (ref=%ld, len=%ld/%ld)}", cell->data.str->str, (long)cell->data.str->refco, (long)cell->data.str->used, (long)cell->data.str->alloced); else printf("%s", cell->data.str->str); return;
		case FAWK_STRNUM: if (verbose) printf("STRNUM:{"FAWK_NUM_PRINTF_FMT" '%s' (ref=%ld, len=%ld/%ld)}", cell->data.str->num, cell->data.str->str, (long)cell->data.str->refco, (long)cell->data.str->used, (long)cell->data.str->alloced); else printf("%s", cell->data.str->str); return;
		case FAWK_ARRAY:  printf("ARRAY:{uid=%ld len=%ld}", cell->data.arr->uid, (long)cell->data.arr->hash.used); return;
		case FAWK_FUNC:   printf("FUNC:{%s}", cell->data.func.name); return;
#warning TODO: refine this
		case FAWK_SYMREF: printf("SYMREF"); return;
		case FAWK_CCALL_RET: printf("CCAL_RET"); return;
		case FAWK_NIL:    printf("NIL"); return;
	}
	printf("<invalid cell>");
}
