
/*
libfawk - A function-only AWK dialect - compacted, single-source-file version
(for a human readable source code please visit the project page)

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS IS AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The regex engine included here is derived from the public domain regex engine
by Ozan S. Yigit (oz). For the human readable version see:
svn://repo.hu/genregex/trunk

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#ifndef fawk_malloc
#define fawk_malloc(ctx,size) malloc(size)
#define fawk_calloc(ctx,n,m) calloc(n, m)
#define fawk_realloc(ctx,ptr,size) realloc(ptr, size)
#define fawk_free(ctx,ptr) free(ptr)
#endif
#ifndef fawk_assert
#include <assert.h>
#define fawk_assert assert
#endif
#ifndef fawk_no_math
#include <math.h>
#define fawk_strtod(str) strtod(str, NULL)
#define fawk_fmod(x,y) fmod(x, y)
#define FAWK_NUM_PRINTF_FMT "%g"
#endif
#define RE_PRFX(name) fawk_re_ ## name
#define re_tolower(c) c
typedef struct RE_PRFX(s) RE_PRFX(t);
typedef enum re_error_e {
 FAWK_RE_ERR_SUCCESS = 0,
FAWK_RE_ERR_NOPREV, FAWK_RE_ERR_NOCLCL, FAWK_RE_ERR_EMPTYCLO, FAWK_RE_ERR_BADCLO, FAWK_RE_ERR_STACK, FAWK_RE_ERR_EMPTYTAG, FAWK_RE_ERR_UNMATCHTAGC, FAWK_RE_ERR_EMPTYWORD, FAWK_RE_ERR_CYCREF, FAWK_RE_ERR_UNTERMREF, FAWK_RE_ERR_UNMATCHTAGO, FAWK_RE_ERR_CLOBADNFA, FAWK_RE_ERR_BADNFA } re_error_t;

#define MAXTAG 10
enum opcode_e {
 NOP = 0,
 END = 0,
 CHR = 1,
 ANY = 2,
 CCL = 3,
 BOL = 4,
 EOL = 5,
 BOT = 6,
 EOT = 7,
 BOW = 8,
 EOW = 9,
 REF = 10,
 CLO = 11
};
#define MAXCHR 128
#define BITBLK MAXCHR/8
#define BLKIND 0170
#define BITIND 07
typedef unsigned char CHAR;
static CHAR bitarr[] = {1,2,4,8,16,32,64,128};
struct RE_PRFX(s) {
 CHAR *nfa;
 int nfa_used, nfa_alloced;
 CHAR bittab[BITBLK];
 const char *bol, *bopat[MAXTAG], *eopat[MAXTAG];
 char *dst;
 int dst_used, dst_alloced;
 re_error_t error;
 char fail_op;
};
static void
chset(RE_PRFX(t) *r, CHAR c)
{
 r->bittab[(CHAR) ((c) & BLKIND) >> 3] |= bitarr[(c) & BITIND];
}
#define badpat(x) do { r->error = x; goto error; } while(0)
#define grow(r,buf,delta) \
 if (r->buf ## _used + delta >= r->buf ## _alloced) { \
  r->buf ## _alloced += 256 + delta; \
  r->buf = realloc(r->buf, r->buf ## _alloced); \
 }
#define store(x) \
 do { \
  grow(r, nfa, 1); \
  *mp++ = x; \
 } while(0)
RE_PRFX(t) *RE_PRFX(comp)(const char *pat)
{
 int tagstk[MAXTAG];
 int sta;
 register const char *p;
 register CHAR *mp;
 register CHAR *lp;
 register CHAR *sp;
 register int tagi = 0;
 register int tagc = 1;
 register int n;
 register CHAR mask;
 int c1, c2;
 RE_PRFX(t) *r;
 r = calloc(1, sizeof(RE_PRFX(t)));
 sta = NOP;
 grow(r, nfa, 1);
 sp = mp = r->nfa;
 if (!pat || !*pat) {
  if (sta)
   return NULL;
  else
   badpat(FAWK_RE_ERR_NOPREV);
 }
 sta = NOP;
 for (p = pat; *p; p++) {
  lp = mp;
  switch(*p) {
  case '.':
   store(ANY);
   break;
  case '^':
   if (p == pat)
    store(BOL);
   else {
    store(CHR);
    store(re_tolower(*p));
   }
   break;
  case '$':
   if (!*(p+1))
    store(EOL);
   else {
    store(CHR);
    store(re_tolower(*p));
   }
   break;
  case '[':
   store(CCL);
   if (*++p == '^') {
    mask = 0377;
    p++;
   }
   else
    mask = 0;
   if (*p == '-')
    chset(r, *p++);
   if (*p == ']')
    chset(r, *p++);
   while (*p && *p != ']') {
    if (*p == '-' && *(p+1) && *(p+1) != ']') {
     p++;
     c1 = *(p-2) + 1;
     c2 = *p++;
     while (c1 <= c2)
      chset(r, re_tolower((CHAR)c1++));
    }
    else if (*p == '\\') {
     p++;
     switch (*p) {
      case 'n': chset(r, '\n'); break;
      case 'r': chset(r, '\r'); break;
      case 't': chset(r, '\t'); break;
      case '\\': chset(r, '\\'); break;
      case '\0': p--; chset(r, '\\'); break;
      default: chset(r, *p);
     }
     p++;
    }
    else
     chset(r, re_tolower(*p++));
   }
   if (!*p)
    badpat(FAWK_RE_ERR_NOCLCL);
   for (n = 0; n < BITBLK; r->bittab[n++] = (char) 0)
    store(mask ^ r->bittab[n]);
   break;
  case '*':
  case '+':
   if (p == pat)
    badpat(FAWK_RE_ERR_EMPTYCLO);
   lp = sp;
   if (*lp == CLO)
    break;
   switch(*lp) {
   case BOL:
   case BOT:
   case EOT:
   case BOW:
   case EOW:
   case REF:
    badpat(FAWK_RE_ERR_BADCLO);
   default:
    break;
   }
   if (*p == '+')
    for (sp = mp; lp < sp; lp++)
     store(*lp);
   store(END);
   store(END);
   sp = mp;
   while (--mp > lp)
    *mp = mp[-1];
   store(CLO);
   mp = sp;
   break;
  case '\\':
   switch(*++p) {
   case '(':
    if (tagc < MAXTAG) {
     tagstk[++tagi] = tagc;
     store(BOT);
     store(tagc++);
    }
    else
     badpat(FAWK_RE_ERR_STACK);
    break;
   case ')':
    if (*sp == BOT)
     badpat(FAWK_RE_ERR_EMPTYTAG);
    if (tagi > 0) {
     store(EOT);
     store(tagstk[tagi--]);
    }
    else
     badpat(FAWK_RE_ERR_UNMATCHTAGC);
    break;
   case '<':
    store(BOW);
    break;
   case '>':
    if (*sp == BOW)
     badpat(FAWK_RE_ERR_EMPTYWORD);
    store(EOW);
    break;
   case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
    n = *p-'0';
    if (tagi > 0 && tagstk[tagi] == n)
     badpat(FAWK_RE_ERR_CYCREF);
    if (tagc > n) {
     store(REF);
     store(n);
    }
    else
     badpat(FAWK_RE_ERR_UNTERMREF);
    break;
   case 'b': store(CHR); store('\b'); break;
   case 'n': store(CHR); store('\n'); break;
   case 'f': store(CHR); store('\f'); break;
   case 'r': store(CHR); store('\r'); break;
   case 't': store(CHR); store('\t'); break;
   default:
    store(CHR);
    store(re_tolower(*p));
   }
   break;
  default :
   store(CHR);
   store(re_tolower(*p));
   break;
  }
  sp = lp;
 }
 if (tagi > 0)
  badpat(FAWK_RE_ERR_UNMATCHTAGO);
 store(END);
error:;
 return r;
}
static const char *pmatch(RE_PRFX(t) *r, const char *, const char *, CHAR *, int *);
#define end(s) (*(s) == '\0')
int RE_PRFX(exec)(RE_PRFX(t) *r, const char *lp)
{
 register CHAR c;
 register const char *ep = 0;
 register CHAR *ap = r->nfa;
 int score = 1;
 const char *lpend = NULL;
 r->bol = lp;
 r->bopat[0] = r->bopat[1] = r->bopat[2] = r->bopat[3] = r->bopat[4] = 0;
 r->bopat[5] = r->bopat[6] = r->bopat[7] = r->bopat[8] = r->bopat[9] = 0;
 r->error = 0; r->fail_op = 0;
 switch(*ap) {
 case BOL:
  ep = pmatch(r, lp, lpend, ap, &score);
  break;
 case CHR:
  c = *(ap+1);
  while (!end(lp) && re_tolower(*lp) != c)
   lp++;
  if (end(lp))
   return 0;
 default:
  do {
   if ((ep = pmatch(r, lp, lpend, ap, &score)))
    break;
  } while (!end(lp++));
  break;
 case END:
  return 0;
 }
 if (!ep)
  return 0;
 r->bopat[0] = lp;
 r->eopat[0] = ep;
 return score;
}
static CHAR chrtyp[MAXCHR] = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };

#define inascii(x) (0177&(x))
#define iswordc(x) chrtyp[inascii(x)]
#define isinset(x,y) ((x)[((y)&BLKIND)>>3] & bitarr[(y)&BITIND])
#define ANYSKIP 2
#define CHRSKIP 3
#define CCLSKIP 18
static const char *pmatch(RE_PRFX(t) *r, const char *lp, const char *lpend, CHAR *ap, int *score)
{
 register int op, c, n;
 register const char *e;
 register const char *bp;
 register const char *ep;
 const char *are;
 while ((op = *ap++) != END)
  switch(op) {
  case CHR:
   if (re_tolower(*lp++) != *ap++)
    return 0;
   (*score) += 100;
   break;
  case ANY:
   if (end(lp++))
    return 0;
   (*score)++;
   break;
  case CCL:
   if (end(lp))
    return 0;
   c = re_tolower(*lp++);
   if (!isinset(ap,c))
    return 0;
   ap += BITBLK;
   (*score) += 2;
   break;
  case BOL:
   if (lp != r->bol)
    return 0;
   (*score) += 10;
   break;
  case EOL:
   if (!end(lp))
    return 0;
   (*score) += 10;
   break;
  case BOT:
   r->bopat[*ap++] = lp;
   break;
  case EOT:
   r->eopat[*ap++] = lp;
   break;
   case BOW:
   if ((lp!=r->bol && iswordc(lp[-1])) || !iswordc(*lp))
    return 0;
   (*score) += 5;
   break;
  case EOW:
   if (lp==r->bol || !iswordc(lp[-1]) || (!end(lp) && iswordc(*lp)))
    return 0;
   (*score) += 5;
   break;
  case REF:
   n = *ap++;
   bp = r->bopat[n];
   ep = r->eopat[n];
   while (bp < ep) {
    if (*bp++ != *lp++)
     return 0;
   (*score) += 2;
   }
   break;
  case CLO:
   are = lp;
   switch(*ap) {
   case ANY:
    while (!end(lp))
     lp++;
    n = ANYSKIP;
    (*score)++;
    break;
   case CHR:
    c = *(ap+1);
    while (!end(lp) && c == re_tolower(*lp))
     lp++;
    n = CHRSKIP;
    (*score) += 100;
    break;
   case CCL:
    while ((c = re_tolower(*lp)) && isinset(ap+1,c))
     lp++;
    n = CCLSKIP;
    (*score) += 2;
    break;
   default:
    r->error = FAWK_RE_ERR_CLOBADNFA;
    r->fail_op = *ap;
    return 0;
   }
   ap += n;
   while (lp >= are) {
    e = pmatch(r, lp, lpend, ap, score);
    if (e)
     return e;
    --lp;
   }
   return 0;
  default:
   r->error = FAWK_RE_ERR_BADNFA;
   r->fail_op = op;
   return 0;
  }
 return lp;
}
static void dappend(RE_PRFX(t) *r,char c)
{
 grow(r, dst, 1);
 r->dst[r->dst_used++] = c;
}
static void dappend_multi(RE_PRFX(t) *r, const char *s, int len)
{
 grow(r, dst, len);
 memcpy(r->dst + r->dst_used, s, len);
 r->dst_used += len;
}
int RE_PRFX(backref)(RE_PRFX(t) *r, char **dst, const char *src)
{
 register char c;
 register int pin;
 register const char *bp, *ep;
 if (dst != NULL) {
  *dst = 0;
  r->dst_used = 0;
 }
 if (!*src || !r->bopat[0])
  return 0;
 while ((c = *src++)) {
  switch(c) {
  case '\\':
   c = *src++;
   if (c >= '0' && c <= '9') {
    pin = c - '0';
    break;
   }
  default:
   dappend(r, c);
   continue;
  }
  if ((bp = r->bopat[pin]) && (ep = r->eopat[pin])) {
   while (*bp && bp < ep)
    dappend(r, *bp++);
   if (bp < ep)
    return -1;
  }
 }
 if (dst != NULL) {
  dappend(r, '\0');
  *dst = r->dst;
 }
 return r->dst_used;
}
int RE_PRFX(subst)(RE_PRFX(t) *r, char **dst, const char *src, const char *pat, int first_only)
{
 const char *s, *s_end, *s_rest;
 int cnt;
 s_rest = s = src;
 for(s_end = s; *s_end != '\0'; s_end++) ;
 cnt = 0;
 r->dst_used = 0;
 for(;;) {
  if (RE_PRFX(exec)(r, s)) {
   cnt++;
   s_rest = r->eopat[0];
   dappend_multi(r, s, r->bopat[0] - s);
   RE_PRFX(backref)(r, NULL, pat);
   if (first_only)
    goto last;
   s = r->eopat[0];
  }
  else {
   goto last;
  }
 }
last:;
  if (s_end - s_rest > 0)
   dappend_multi(r, s_rest, s_end - s_rest);
  dappend(r, '\0');
  *dst = r->dst;
  return cnt;
}
void RE_PRFX(free)(RE_PRFX(t) *re)
{
 if (re->dst != NULL)
  free(re->dst);
 if (re->nfa != NULL)
  free(re->nfa);
 free(re);
}
#undef RE_PRFX
static unsigned fawk_re_ptrhash(const void *k) {
 const unsigned long n = (const unsigned long)k;
 return (n >> 2) ^ (n >> 12);
}
static int fawk_re_ptrkeyeq(const void *a, const void *b) { return a == b; }
static void regexp_str_free_cb(fawk_pkg_t *pkg, fawk_ctx_t *ctx, fawk_str_t *str)
{
 fawk_htpp_t *cache = pkg->data[0].ptr;
 fawk_re_t *re = fawk_htpp_pop(cache, str);
 if (re != NULL)
  fawk_re_free(re);
}
static void regexp_uninit_cb(fawk_pkg_t *pkg, fawk_ctx_t *ctx)
{
 fawk_htpp_t *cache = pkg->data[0].ptr;
 fawk_htpp_uninit(cache);
 fawk_free(ctx, cache);
}
static fawk_htpp_t *get_cache(fawk_ctx_t *ctx)
{
 fawk_pkg_t *pkg;
 for(pkg = ctx->pkg_head; pkg != NULL; pkg = pkg->next)
  if (pkg->uninit_cb == regexp_uninit_cb)
   return pkg->data[0].ptr;
 return NULL;
}
#define cache_str(S) \
 re = fawk_htpp_get(cache, S->data.str); \
 if (re == NULL) fawk_htpp_set(cache, S->data.str, re = fawk_re_comp(S->data.str->str));
#define set_arr_str(arr,idx,string,string_len) \
 do { \
  fawk_cell_t *elem = fawk_array_resolve_c(ctx, 1, arr, FAWK_NUM, (fawk_num_t)idx, FAWK_NIL); \
  if (string_len != 0) { elem->type = FAWK_STR; elem->data.str = fawk_str_new_from_literal(ctx, string, string_len); } \
  else elem->type = FAWK_NIL; \
 } while(0)
static void fawk_regexp_match(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
 fawk_htpp_t *cache = get_cache(ctx);
 fawk_re_t *re;
 fawk_cell_t *opat = NULL, *oaddr = NULL;
 int n;
 if ((argc < 2) || (argc > 4) || (FAWK_CFUNC_ARG(1)->type != FAWK_STR))
  return;
 if (fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(0)) != 0)
  return;
 if (argc > 2) {
  opat = FAWK_CFUNC_ARG(2);
  if (opat->type != FAWK_ARRAY)
   return;
 }
 if (argc > 3) {
  oaddr = FAWK_CFUNC_ARG(3);
  if (oaddr->type != FAWK_ARRAY)
   return;
 }
 cache_str(FAWK_CFUNC_ARG(1));
 retval->type = FAWK_NUM;
 retval->data.num = fawk_re_exec(re, FAWK_CFUNC_ARG(0)->data.str->str);
 if (opat != NULL) {
  for(n = 0; n < MAXTAG; n++) {
   if (re->bopat[n] != NULL) {
    set_arr_str(opat, n, re->bopat[n], re->eopat[n]-re->bopat[n]);
   }
  }
 }
 if (oaddr != NULL) {
  for(n = 0; n < MAXTAG; n++) {
   if (re->bopat[n] != NULL) {
    fawk_cell_t *elem = fawk_array_resolve_c(ctx, 1, oaddr, FAWK_NUM, (fawk_num_t)(n*2), FAWK_NIL);
    elem->type = FAWK_NUM; elem->data.num = re->bopat[n] - re->bol;
    elem = fawk_array_resolve_c(ctx, 1, oaddr, FAWK_NUM, (fawk_num_t)(n*2+1), FAWK_NIL);
    elem->type = FAWK_NUM; elem->data.num = re->eopat[n] - re->bopat[n];
   }
  }
 }
}
static void fawk_regexp_sub(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
 fawk_htpp_t *cache = get_cache(ctx);
 fawk_re_t *re;
 fawk_cell_t *str, *v, *arr;
 int first_only;
 char *dst;
 if (argc != 3)
  return;
 if ((fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(0)) != 0) || (fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(1)) != 0))
  return;
 first_only = (fname[0] != 'g');
 cache_str(FAWK_CFUNC_ARG(0));
 v = FAWK_CFUNC_ARG(2);
 if (v->type == FAWK_SYMREF) {
  if (((str = fawk_symtab_deref(ctx, &v->data.symref, 0, &arr)) == NULL) || ((str->type != FAWK_STR) && (str->type != FAWK_STRNUM)))
   return;
  retval->type = FAWK_NUM;
  retval->data.num = fawk_re_subst(re, &dst, str->data.str->str, FAWK_CFUNC_ARG(1)->data.str->str, first_only);
  fawk_str_free(ctx, str->data.str);
  str->type = FAWK_STR;
  str->data.str = fawk_str_new_from_literal(ctx, dst, -1);
 }
 else {
  if (fawk_cast_to_str(ctx, v) != 0)
   return;
  fawk_re_subst(re, &dst, v->data.str->str, FAWK_CFUNC_ARG(1)->data.str->str, first_only);
  retval->type = FAWK_STR;
  retval->data.str = fawk_str_new_from_literal(ctx, dst, -1);
 }
}
static void fawk_regexp_split(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
 fawk_htpp_t *cache = get_cache(ctx);
 fawk_re_t *re;
 fawk_cell_t *arr;
 const char *s;
 long n;
 if (argc != 3)
  return;
 if ((fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(0)) != 0) || (FAWK_CFUNC_ARG(1)->type != FAWK_ARRAY) || (fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(2)) != 0))
  return;
 s = FAWK_CFUNC_ARG(0)->data.str->str;
 arr = FAWK_CFUNC_ARG(1);
 if (FAWK_CFUNC_ARG(2)->data.str->str[0] == '\0') {
  for(n = (fname[5] != '0'); *s != '\0'; n++,s++)
   set_arr_str(arr, n, s, 1);
 }
 else {
  cache_str(FAWK_CFUNC_ARG(2));
  for(n = (fname[5] != '0'); (*s != '\0') && (fawk_re_exec(re, s) > 0); n++) {
   if (re->eopat[0] == re->bol) { re->bopat[0]++; re->eopat[0]++; }
   set_arr_str(arr, n, s, re->bopat[0]-re->bol);
   s = re->eopat[0];
  }
  set_arr_str(arr, n, s, -1);
  n++;
 }
 retval->type = FAWK_NUM; retval->data.num = n - (fname[5] != '0');
}
FAWK_API int fawk_regexp_init(fawk_ctx_t *ctx)
{
 fawk_pkg_t *pkg;
 fawk_htpp_t *cache;
 pkg = fawk_calloc(ctx, sizeof(fawk_pkg_t), 1);
 if (pkg == NULL) return -1;
 pkg->next = ctx->pkg_head;
 ctx->pkg_head = pkg;
 pkg->str_free_cb = regexp_str_free_cb;
 pkg->uninit_cb = regexp_uninit_cb;
 fawk_htpp_init(pkg->data[0].ptr = cache = fawk_malloc(ctx, sizeof(fawk_htpp_t)), fawk_re_ptrhash, fawk_re_ptrkeyeq);
 if (cache == NULL) {
  free(pkg);
  return -1;
 }
 fawk_symtab_regcfunc(ctx, "match", fawk_regexp_match);
 fawk_symtab_regcfunc(ctx, "sub", fawk_regexp_sub);
 fawk_symtab_regcfunc(ctx, "gsub", fawk_regexp_sub);
 fawk_symtab_regcfunc(ctx, "split", fawk_regexp_split);
 fawk_symtab_regcfunc(ctx, "split0", fawk_regexp_split);
 return 0;
}
